import logging
import mimetypes
import os
import re
import sys

from bn import AttributeDict, uniform_path, absimport
from flows.provider import ServiceProvider, start_flow
from httpkit.service.http import HTTPObject

log = logging.getLogger(__name__)

#
# Other flows apps
#

def run_cgi(flow, resource, executable):
    from cgipack.cgiapp import CGIApplication
    CGIApplication(
        script=resource, 
        executable=executable,
    )(flow)

def run_script(flow, resource):
    try:
        execfile(resource, {"__name__": "__main__", 'flow': flow})
    except SystemExit, sts:
        log.error("CGI script exit status %s", str(sts))
    else:
        log.info("CGI script exited OK")

def not_found(flow, emit_log=True):
    if emit_log:
        log.error('Not found: %s'%flow.http.environ['PATH_INFO'])
    flow.http.response.status = '404 Not Found'
    if not flow.has_key('error_document'):
        try:
            flow.provider.start('error_document')
        except:
            flow.http.response.headers = [('Content-Type', 'text/plain')]
            flow.http.response.body = [u'Not found']
            return
    flow.error_document.render()

def forbidden(flow, emit_log=True):
    if emit_log:
        log.error('Forbidden: %s'%flow.http.environ['PATH_INFO'])
    if not flow.has_key('error_document'):
        flow.provider.start('error_document')
    flow.http.response.status = '403 Forbidden'
    if not flow.has_key('error_document'):
        try:
            flow.provider.start('error_document')
        except:
            flow.http.response.headers = [('Content-Type', 'text/plain')]
            flow.http.response.body = [u'Not found']
            return
    flow.error_document.render()

#
# WSGI Middleware Adaptor
#

class SetUpFlow(object):
    def __init__(self, flows_application, provider, start=[]):
        self.provider = provider
        self.application = flows_application
        self.start = start

    def __call__(self, environ, start_response):
        existing = {
            'app':  AttributeDict(
                try_next = True,
            )
        }
        existing['http'] = AttributeDict(response=HTTPObject(existing))
        existing['http']['environ'] = environ

        result = start_flow(
            self.application,
            provider = self.provider,
            required_services=self.start,
            existing=existing,
        )

        start_response(
            existing['http'].response.status, 
            existing['http'].response.headers, 
        )
        return existing['http'].response.get_result()

def make_application(flow, start=['error_report'], extra_components=[]):
    available_services = absimport(
        flow.config.app.package+'.framework.service'
    ).on_load_services(flow)
    application = SetUpFlow(
        make_flows_application(flow, extra_components), 
        provider = ServiceProvider(
            existing_services=dict(config=flow.config), 
            available_services=available_services,
        ), 
        start=start,
    )
    return application

def staticDispatch(
    site_root, 
    directory_index=None, 
    enable_cgi=False, 
    enable_script=False, 
    default_content_type=u'application/octet-stream',
    config = None,
):
    site_root = uniform_path(site_root)+os.sep
    config = config or {}

    def static_application(flow):
        # Make sure the root path ends with a /
        path_info = flow.http.environ.get('PATH_INFO', '')
        if not path_info:
            # There is a bug somewhere, let's try to work around it
            log.error('Empty PATH_INFO when trying to do a static dispatch')
            path_info = '/'
        resource = uniform_path(os.path.join(site_root, path_info[1:]))
        if os.path.exists(resource) and os.path.isdir(resource) and directory_index is not None:
            resource = os.path.join(resource, directory_index)
        if not os.path.exists(resource):
            return not_found(flow, emit_log=False)
        if not resource.startswith(site_root):
            # Trying to get a file it shouldn't have access to
            return not_found(flow)
        # Deal with CGI scripts
        if resource.endswith('.py'):
            if enable_cgi:
                flow['app']['resource'] = resource
            return not_found(flow)
            #run_cgi(flow, resource)
        elif resource.endswith('.flo'):
            if enable_script:
                flow['app']['resource'] = resource
            return not_found(flow)
            #run_script(flow, resource)
        else:
            # Deal with static files
            type, encoding = mimetypes.guess_type(resource)
            if not type:
                type = default_content_type
            try:
                fp = open(resource, 'rb')
            except (IOError, OSError), e:
                return forbidden(flow)
            else:
                def result_iterator():
                    try:
                        for line in fp:
                            if line:
                                yield line
                    finally:
                        fp.close()
                flow.http.response.body = result_iterator()
                flow.http.response.status = '200 OK'
                flow.http.response.headers = [('Content-Type', type)]
    return static_application

def componentDispatch(
    map,
    cgi_executable=None,
    enable_cgi=False,
    enable_script=False,
    enable_static=False,
):
    components = {}
    def component_application(flow):
        # Special handling for ``.py`` and ``.flo`` files
        path_info = flow.http.environ.get('PATH_INFO', '')
        resource = None
        if enable_cgi or enable_script:
            if enable_static:
                # The static_app has already looked for a resource,
                # no need to look for it again
                if flow.app.has_key('resource'):
                    resource = flow.app.resource
            else:
                # The static server hasn't looked for it. We better check 
                # instead
                site_root = uniform_path(
                    os.path.join(flow.config.app.app_dir, 'static')
                )
                site_root = uniform_path(site_root)+'/'
                resource = uniform_path(os.path.join(site_root, path_info[1:]))
        if resource and (resource[-3:] == '.py' or resource[-4:] == '.flo'):
            if not os.path.exists(resource):
                return not_found(flow)
            if resource[-3:] == '.py':
                if enable_cgi:
                    return run_cgi(flow, resource, cgi_executable)
            elif resource[-4:] == '.flo':
                if enable_script:
                    return run_script(flow, resource)
        request_url = flow.http.environ['PATH_INFO']
        mod = None
        for expr, value in map:
            match = re.match(expr, flow.http.environ['PATH_INFO'])
            if match:
                mod = value
                break
        if not mod:
            log.warning('No such module for %r', flow.http.environ['PATH_INFO'])
            return not_found(flow)
        if not components.has_key(mod):
            log.info('Loading component %r', mod)
            components[mod] = mod(flow)
        result = components[mod](flow)
        if result:
            flow.http.response.body.append(result)
        # The result will be present as flow.wgsi.response
    return component_application

def simpleCascade(static, component):
    def simpleCascade_application(flow):
        static(flow)
        if flow.http.response.status[:3] == '404' and flow.app.try_next:
            # Reset HTTP
            flow.http.response.status = '200 OK'
            flow.http.response.headers = [
                ('Content-Type', 'text/html; charset=utf-8')
            ]
            flow.http.response.body = []
            component(flow)
    return simpleCascade_application

def make_flows_application(flow, extra_components=[]):
    map = extra_components[:]
    for item in flow.config.app.auto_connect:
        map.append(
            (
                '^/%s/(.*)'%item, 
                absimport(
                    flow.config.app.package+'.ext.'+item+'.dispatch'
                ).on_load_component
            )
        )
    map.append(
        (
            '^(.*)$', 
            absimport(
                flow.config.app.package+'.dynamic.dispatch'
            ).on_load_component
        )
    )
    component = componentDispatch(
        map=map,
        cgi_executable=flow.config.app.cgi_executable,
        enable_cgi=flow.config.app.enable_cgi,
        enable_script=flow.config.app.enable_script,
        enable_static=flow.config.app.enable_static,
    )
    #flow.config.app['url_map'] = map
    if flow.config.app.enable_static:
        static = staticDispatch(
            site_root = uniform_path(
                os.path.join(flow.config.app.app_dir, 'static')
            ),
            directory_index = 'index.html',
            enable_cgi = flow.config.app.enable_cgi,
            enable_script = flow.config.app.enable_script,
            default_content_type = \
               flow.config.app.static_default_content_type,
            config = flow.config,
        )
        return simpleCascade(static, component)
    else:
        return component

