import getopt
import os
import logging
import logging.config
import warnings

from bn import absimport, AttributeDict, uniform_path
from configconvert import existingDirectory, existingFile, parse_config,\
   handle_option_error, handle_section_error, split_options
from conversionkit import Conversion, chainConverters, noConversion
from recordconvert import toRecord
from stringconvert import unicodeToInteger
from stringconvert import unicodeToFloat
from stringconvert import unicodeToBoolean

#
# Custom converters
#

def unicodeSplit(separator=','):
    def unicodeSplit_converter(conversion, state):
        if not isinstance(conversion.value, (unicode)):
            raise Exception('Expected the input to be Unicode string')
        conversion.result = [x.strip() for x in conversion.value.split(separator)]
    return unicodeSplit_converter

def checkContentType():
    def checkContentType_converter(conversion, state):
        if not isinstance(conversion.value, (unicode)):
            raise Exception('Expected the input to be Unicode string, not %r'%conversion.value)
        if not len(conversion.value.split('/')) == 2:
            raise Exception(
                "Invalid default content type. Expected the format "
                "'text/html'"
            )
        conversion.result = unicode(conversion.value)
    return checkContentType_converter

#
# Helpers
#

def create_config(config_file, logging=True):
    if not os.path.exists(config_file):
        raise getopt.GetoptError('No such file %r'%config_file)
    options = parse_config(config_file)
    if not options.has_key('app.package'):
        raise getopt.GetoptError(
            "No 'app.package' option found in config file specified"
        )
    # Set some key parts and move to the correct directory
    set_up_app_options(options, config_file)
    # Load the config object
    config = absimport(
        options['app.package']+'.framework.config',
    ).on_load_config(options)
    if logging:
        set_up_logging(config)
    return config

def set_up_logging(config, logging_file=None):
    if logging_file is None:
        if config.option_group.has_key('logging'):
            logging_file = config.option_group['logging'].get('filename')
    if logging_file is None:
        logging_file = config.app.config_file[:-5]+'.logging'
    if not os.path.exists(logging_file):
        raise getopt.GetoptError(
            "Could not find logging configutaion file %r"%logging_file
        )
    logging.config.fileConfig(logging_file)
    logging_handlers = []
    in_loggers = False
    fp = open(logging_file)
    for line in fp:
        if line.startswith('[loggers]'):
            in_loggers = True
        if in_loggers and line.startswith('keys') and '=' in line:
            for handler in line[line.find('=')+1:].strip().split(','):
                logging_handlers.append(handler.strip())
            break
    fp.close()
    if not 'flows' in logging_handlers:
        warnings.warn(
            "The 'flows' module does not have a logger set up"
        )
    config['logging'] = AttributeDict(handlers=logging_handlers)

def set_up_app_options(options, config_file):
    # Set the app settings
    for auto_generated in [
        'app.app_dir', 
        'app.deploy_dir', 
        'app.config_file'
    ]:
        if options.has_key(auto_generated):
            raise Exception(
                'The option %r is auto-generated and should not appear in '
                'the config file'%(auto_generated,)
            )
    options['app.app_dir'] = __import__(options['app.package']).__path__[0]
    options['app.config_file'] = uniform_path(config_file)
    options['app.deploy_dir'] = os.path.split(options['app.config_file'])[0]
    os.chdir(options['app.deploy_dir'])

def parse_and_validate_options(options):
    """\
    This function is where you can set up your settings. Ordinarily, 
    settings which you want end users to configure should go in the 
    config file, settings which the developer needs control over can
    go here.
    """

    # Split up the setting into parts
    config = AttributeDict()
    config['option'] = options
    option_group = split_options(options)
    config['option_group'] = option_group
    if config.option_group.has_key('ext'):
        raise Exception(
            'The config file cannot have any options starting '
            '``ext.``'
        ) 
    for option, example in [
        ('app', "'app.package'"), 
        ('server', "'server.host'"),
    ]:
        if not option_group.has_key(option):
            handle_section_error(None, option, example)
    # Application options
    app_converter = chainConverters(
        toRecord(
            missing_errors =  [
                "The required option 'app.%(key)s' is missing", 
                (
                    'package', 
                    #'site_packages', 
                    'debug',
                    'app_dir',
                    'config_file',
                    'deploy_dir',
                )
            ],
            empty_errors = [
                "The option 'app.%(key)s' cannot be empty", 
                (
                    'package', 
                    #'site_packages', 
                    'debug', 
                    'app_dir',
                    'config_file',
                )
            ],
            missing_defaults = {
                'auto_connect': u'', 
                'enable_cgi': False,
                'enable_script': False,
                'enable_static': False,
                'static_default_content_type': u'application/octet-stream',
                'cgi_executable': None,
            },
            converters=dict(
                package = noConversion(),
                #site_packages = noConversion(),
                cgi_executable = noConversion(),
                debug = unicodeToBoolean(),
                static_default_content_type = checkContentType(),
                app_dir = existingDirectory(),
                deploy_dir = existingDirectory(),
                config_file = existingFile(),
                auto_connect = unicodeSplit(),
                enable_cgi = unicodeToBoolean(),
                enable_script = unicodeToBoolean(),
                enable_static = unicodeToBoolean(),
            ),
            allow_extra_fields=False,
            msg_field_not_allowed=\
               "The config file does not support the 'app.%s' option.",
            msg_fields_not_allowed=\
               "The 'app.*' options in the config file are invalid. The "
               "options '%s' and '%s' are not supported.",
        ),
    )
    conversion = Conversion(option_group['app']).perform(app_converter)
    if not conversion.successful:
        handle_option_error(conversion, 'app')
    else:
        config['app'] = conversion.result
    # Server options
    server_converters={
        'host': noConversion(),
        'port': unicodeToInteger(),
        'profile': unicodeToBoolean(),
        'display_host': noConversion(),
        'display_port': unicodeToInteger(),
    }
    server_converter = chainConverters(
        toRecord(
            missing_errors=[
                "The required option 'server.%(key)s' is missing", 
                ('display_host', 'display_port'),
            ],
            missing_defaults={
                "profile": False,
            },
            empty_errors="The option 'server.%(key)s' cannot be empty",
            converters=server_converters,
        ),
    )
    conversion = Conversion(option_group['server']).perform(
        server_converter
    )
    if not conversion.successful:
        handle_option_error(conversion, 'server')
    else:
        config['server'] = conversion.result
    return config

