"""\
Flow provider service
"""

# It may actually turn out to be faster to have an events based system rather
# than checking for the presence of ``start()``, ``stop()`` and ``error()``
# for each one? Needs testing.

import logging
import os
import warnings

from bn import AttributeDict, uniform_path, absimport

log = logging.getLogger(__name__)

class ServiceError(Exception):
    pass
#
# Flow Provider
#

class ServiceProvider(object):
    def __init__(
        self, 
        existing_services=None, 
        available_services=None, 
        name='provider', 
        check_config=True
    ):
        self.instances = {}
        self.existing_services = existing_services or {}
        self.available_services = available_services or {}
        self.available_configs = {}
        self.name = name
        if check_config and self.available_services:
            if existing_services:
                flow = AttributeDict(existing_services)
            else:
                flow = None
            for name in self.available_services.keys():
                if hasattr(self.available_services[name], 'config'):
                    self.available_configs[name] = \
                       self.available_services[name].config(flow, name)

    # environ, config, etc.
    def create(self):
        existing_services = self.existing_services.copy()
        available_services = self.available_services.copy()
        flow = AttributeDict(existing_services)

        def start(name, raise_if_started=True):
            if raise_if_started and hasattr(flow, name):
                raise ServiceError('The %r service is already present'%name)
            if not hasattr(flow, name):
                if not self.available_services.has_key(name):
                    raise ServiceError(
                        'No such service %r. Available: %s'%(
                            name, 
                            repr(available_services.keys())[1:-1] or 'None',

                        )
                    )
                if flow.config.has_key('logging') and \
                   name not in flow.config.logging.handlers:
                    warnings.warn(
                        'The %r service does not have a logger set up' % name
                    )
                if not self.instances.has_key(name):
                    factory = self.available_services[name]
                    if self.available_configs.has_key(name):
                        # Create it from the .create() static method
                        if hasattr(factory, 'create'):
                            service = factory.create(
                                flow, 
                                name, 
                                self.available_configs[name]
                            )
                        else:
                            raise ServiceError(
                                'The service %r does not have a create() '
                                'static method'%name
                            )
                    else:
                        # It will have to parse the config itself
                        service = factory(flow, name)
                    self.instances[name] = service
                if hasattr(self.instances[name], 'start'):
                    self.instances[name].start(flow, name)
                flow.provider.started.insert(0, name)

        def stop(name):
            if hasattr(self.instances[name], 'stop'):
                self.instances[name].stop(flow, name)
            if hasattr(flow, name):
                del flow[name] 

        def error(name):
            if hasattr(self.instances[name], 'error'):
                log.error('Calling the error method of %r', name)
                self.instances[name].error(flow, name)
            if hasattr(flow, name):
                del flow[name]

        def register(new_services, raise_on_error=False):
            for name, service in new_services.items():
                existing_service = None
                if self.available_services.has_key(name):
                    existing_service = self.available_services[name]
                elif self.existing_services.has_key(name):
                    existing_service = self.existing_services[name]
                if existing_service is not None:
                    if service != existing_service:
                        raise ServiceError(
                            'A different service named %r has already been '
                            'registered'%name
                        )
                    else:
                        # It is the same service, just log an error
                        log.error(
                            'The service %r has already been registered as '
                            '%r',
                            name, 
                            service
                        )
                else:
                    self.available_services[name] = service
                    if hasattr(service, 'config'):
                        self.available_configs[name] = service.config(
                            flow,
                            name
                        )
        flow[self.name] = AttributeDict()
        flow.provider['start'] = start
        flow.provider['instance'] = self
        flow.provider['register'] = register
        flow.provider['stop'] = stop
        flow.provider['error'] = error
        flow.provider['error_handled'] = False
        flow.provider['started'] = []
        return flow

def start_flow(
    run, 
    provider,
    existing=None, 
    required_services=[],
):
    """\
    Create a new flow object with the provider set up and with the required
    services started then call the ``run()`` function with the flow as the
    only argument. Handle any errors and stop any services at the end of
    the flow.
    """
    flow = provider.create()
    if existing:
        flow.update(existing)
    for name in required_services:
        flow.provider.start(name)
    reversed = required_services[:]
    reversed.reverse()
    try:
        result = run(flow)
    except:
        for name in reversed:
            flow.provider.error(name)
        if not flow.provider.error_handled:
            raise
    else:
        for name in reversed:
            flow.provider.stop(name)
        return result

def start_service(name):
    """\
    Decorator for starting a service
    """
    def a(func):
        def start_service_(flow):
            try:
                flow.provider.start(name)
            except ServiceError, e:
                raise ServiceError(
                    "@start_service() decorating '%s()' failed with: %r"%(
                        func.__name__, 
                        str(e)
                    )
                )
            return func(flow)
        return start_service_
    return a

